#include <stdio.h>
#include <string.h>

/* Command line program to change tags on mp3 files.
 *
 * Written by:  Chris Studholme
 * Last Update: 11-Jan-1999
 * Copyright:   GPL (http://www.fsf.org/copyleft/gpl.html)
 *
 * Run without parameters to see usage summary.
 * Run with only filename to view mp3 tags.
 * Run with additional parameters to set fields.
 */

typedef struct {
  char tag[3];
  char title[30];
  char artist[30];
  char album[30];
  char year[4];
  char comment[30];
  unsigned char genre;
} tag;


void usage() {
  fprintf(stderr,"\n");

  fprintf(stderr,"NAME\n\tchmp3tag - view/modify tag on MP3 files\n\n");

  fprintf(stderr,"SYNOPSIS\n\tchmp3tag [ options ... ] file\n\n");

  fprintf(stderr,"OPTIONS\n");
  fprintf(stderr,"\t-t title\n\t\tchange song title\n");
  fprintf(stderr,"\t-a artist\n\t\tchange artist\n");
  fprintf(stderr,"\t-l album\n\t\tchange album title\n");
  fprintf(stderr,"\t-y year\n\t\tchange year\n");
  fprintf(stderr,"\t-c comment\n\t\tchange comment\n");
  fprintf(stderr,"\t-q\n\t\tbe quite\n");
  fprintf(stderr,"\n");
}


void cleanstring(char* s) {
  while(*s!=0) 
    if ((*s<32)||(*s>126))
      *s=0;
    else
      ++s;
}


void cleannumber(char* s) {
  while(*s!=0) 
    if ((*s<'0')||(*s>'9'))
      *s=0;
    else
      ++s;
}


/* converts null terminated to fixed width */
char* block(char* dest, const char* src, size_t n) {
  strncpy(dest,src,n);
  for (size_t i=strlen(src); i<n; ++i)
    dest[i]=' ';
  return dest;
}


/* converts fixed width to null terminated */
char* unblock(char* dest, const char* src, size_t n) {
  strncpy(dest,src,n);
  dest[n]=0;
  while((n>0)&&(dest[n-1]==' ')) 
    dest[--n]=0;
  return dest;
}


main(int argc, char*argv[]) {
  if (argc<=1) {
    usage();
    return 1;
  }

  char* newtitle=0;
  char* newartist=0;
  char* newalbum=0;
  char* newyear=0;
  char* newcomment=0;
  char* filename=0;
  bool quite=false;
  bool readonly=true;

  --argc;
  ++argv;
  while (argc>0) {
    if ((strcmp(*argv,"-t")==0)&&(argc>1)) {
      --argc; ++argv;
      newtitle=*argv;
      cleanstring(newtitle);
      if (strlen(newtitle)>30)
	newtitle[30]=0;
      readonly=false;
    }
    else if ((strcmp(*argv,"-a")==0)&&(argc>1)) {
      --argc; ++argv;
      newartist=*argv; 
      cleanstring(newartist);
      if (strlen(newartist)>30)
	newartist[30]=0;
      readonly=false;
    }
    else if ((strcmp(*argv,"-l")==0)&&(argc>1)) {
      --argc; ++argv;
      newalbum=*argv;
      cleanstring(newalbum);
      if (strlen(newalbum)>30)
	newalbum[30]=0;
      readonly=false;
    }
    else if ((strcmp(*argv,"-c")==0)&&(argc>1)) {
      --argc; ++argv;
      newcomment=*argv;
      cleanstring(newcomment);
      if (strlen(newcomment)>30)
	newcomment[30]=0;
      readonly=false;
    }
    else if ((strcmp(*argv,"-y")==0)&&(argc>1)) {
      --argc; ++argv;
      newyear=*argv;
      cleannumber(newyear);
      if (strlen(newyear)>4)
	newyear[4]=0;
      if (strlen(newyear)==2) {
	newyear[2]=newyear[0];
	newyear[3]=newyear[1];
	newyear[0]='1';
	newyear[1]='9';
	newyear[4]=0;
      }
      readonly=false;
    }
    else if (strcmp(*argv,"-q")==0) {
      quite=true;
    }

    else {   /* file argument */
      if (filename!=0) {
	fprintf(stderr,"chmp3tag: only one file at time\n");
	return 1;
      }
      filename=*argv;
    }

    --argc; ++argv;
  }

  if (filename==0) {
    usage();
    return 1;
  }

  // open the mp3 file
  FILE* mp3file = fopen(filename,readonly?"rb":"r+b");
  if (mp3file==NULL) {
    fprintf(stderr,"chmp3tag: failed to open file '%s'\n",filename);
    return 1;
  }

  // find existing tag
  tag newtag;
  bool isNewTag=false;
  if (fseek(mp3file,-sizeof(tag),SEEK_END)!=0) {
    fprintf(stderr,"chmp3tag: failed to seek in file\n");
    fclose(mp3file);
    return 1;
  }
  if (fread(&newtag,sizeof(tag),1,mp3file)!=1) {
    fprintf(stderr,"chmp3tag: failed to read from file\n");
    fclose(mp3file);
    return 1;
  }
  if (strncmp(newtag.tag,"TAG",3)!=0) {
    memset(&newtag,' ',sizeof(tag));
    strncpy(newtag.tag,"TAG",3);
    newtag.genre=0xFF;
    isNewTag=true;
  }

  // modify tag
  if (newartist!=0)
    block(newtag.artist,newartist,30);
  if (newtitle!=0)
    block(newtag.title,newtitle,30);
  if (newalbum!=0)
    block(newtag.album,newalbum,30);
  if (newcomment!=0)
    block(newtag.comment,newcomment,30);
  if (newyear!=0)
    block(newtag.year,newyear,4);

  // display tag
  char buf[31];
  if (!quite) {
    printf("File:    %s\n",filename);
    printf("Title:   %s\n",unblock(buf,newtag.title,30));
    printf("Artist:  %s\n",unblock(buf,newtag.artist,30));
    printf("Album:   %s\n",unblock(buf,newtag.album,30));
    printf("Year:    %s\n",unblock(buf,newtag.year,4));
    printf("Comment: %s\n",unblock(buf,newtag.comment,30));
  }

  // write new tag
  if (!readonly) {
    if (!isNewTag) {
      if (fseek(mp3file,-sizeof(tag),SEEK_END)!=0) {
	fprintf(stderr,"chmp3tag: failed to seek in file (tag not written)\n");
	fclose(mp3file);
	return 1;
      }
    }
    if (fwrite(&newtag,sizeof(tag),1,mp3file)!=1) {
      fprintf(stderr,"chmp3tag: failed to write tag to file\n");
      fclose(mp3file);
      return 1;
    }
  }

  // close the mp3 file and exit
  fclose(mp3file);
  return 0;
}


